﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Drawing;
using System.Drawing.Drawing2D;

namespace HexViewer
{
  /// <summary>HexMap描画クラス</summary>
  public class HexMapDrawer
  {

    #region インスタンス変数・プロパティ

    private Point[,] hexCenters;

    private Point[,][] hexagons;

    private Brush[,] hexBrushes;
    
    /// <summary>Canvas幅を取得する</summary>
    public int Width { get; private set; }

    /// <summary>Canvas高さを取得する</summary>
    public int Height { get; private set; }

    /// <summary>X座標Hex数を取得する</summary>
    public int XNumber { get; private set; }

    /// <summary>Y座標Hex数を取得する</summary>
    public int YNumber { get; private set; }

    /// <summary>Hexの高さを取得する</summary>
    public int HexagonHeight { get; private set; }

    /// <summary>縁取りペンを設定・取得する</summary>
    public Pen PathPen { get; set; } = Pens.Gray;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="xNumber">X軸Hex数</param>
    /// <param name="yNumber">Y軸Hex数</param>
    /// <param name="hexHeight">Hexagonの高さ（内接円の直径）</param>
    public HexMapDrawer(int xNumber, int yNumber, int hexHeight)
    {
      this.XNumber = xNumber;
      this.YNumber = yNumber;

      hexagons = new Point[xNumber, yNumber][];
      hexBrushes = new Brush[xNumber, yNumber];
      hexCenters = new Point[xNumber, yNumber];
      for (int i = 0; i < xNumber; i++)
        for (int j = 0; j < yNumber; j++)
          hexagons[i, j] = new Point[7];

      InitCanvasSize(hexHeight);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>サイズを更新する</summary>
    /// <param name="hexHeight">Hexagonの高さ（内接円の直径）</param>
    public void InitCanvasSize(int hexHeight)
    {
      HexagonHeight = hexHeight;
      int xNumber = hexagons.GetLength(0);
      int yNumber = hexagons.GetLength(1);

      int hit = hexHeight / 2;
      int hit2 = hit + hit;
      int wid = (int)(hit / Math.Tan(Math.PI / 3d));
      int wid2 = wid + wid;
      int wid6 = 6 * wid;

      for (int i = 0; i < xNumber; i++)
      {
        if (i == 0) hexagons[i, 0][0] = new Point(wid, 0);
        else if (i == 1) hexagons[i, 0][0] = new Point(4 * wid, hit);
        else hexagons[i, 0][0] = new Point(hexagons[i - 2, 0][0].X + wid6, hexagons[i - 2, 0][0].Y);

        for (int j = 1; j < yNumber; j++)
          hexagons[i, j][0] = new Point(hexagons[i, j - 1][0].X, hexagons[i, j - 1][0].Y + hit2);
      }

      for (int i = 0; i < xNumber; i++)
      {
        for (int j = 0; j < yNumber; j++)
        {
          hexagons[i, j][1] = new Point(hexagons[i, j][0].X + wid2, hexagons[i, j][0].Y);
          hexagons[i, j][2] = new Point(hexagons[i, j][1].X + wid, hexagons[i, j][1].Y + hit);
          hexagons[i, j][3] = new Point(hexagons[i, j][1].X, hexagons[i, j][2].Y + hit);
          hexagons[i, j][4] = new Point(hexagons[i, j][0].X, hexagons[i, j][3].Y);
          hexagons[i, j][5] = new Point(hexagons[i, j][0].X - wid, hexagons[i, j][2].Y);
          hexagons[i, j][6] = hexagons[i, j][0];
          hexCenters[i, j] = new Point(hexagons[i, j][0].X + wid, hexagons[i, j][5].Y);
        }        
      }

      Width = hexagons[hexagons.GetLength(0) - 1, hexagons.GetLength(1) - 1][2].X;
      if (1 < hexagons.GetLength(0)) Height = hexagons[1, hexagons.GetLength(1) - 1][3].Y + 1;
      else Height = hexagons[0, hexagons.GetLength(1) - 1][3].Y + 1;
    }

    /// <summary>HexMapを描画する</summary>
    /// <param name="grph"></param>
    public void DrawMap(Graphics grph)
    {
      //縁取り線
      GraphicsPath linePath = new GraphicsPath();      

      for (int i = 0; i < hexagons.GetLength(0); i++)
      {
        for (int j = 0; j < hexagons.GetLength(1); j++)
        {
          //縁取り線追加
          linePath.AddPolygon(hexagons[i, j]);

          //内部を塗りつぶし
          GraphicsPath fillPath = new GraphicsPath();
          fillPath.AddPolygon(hexagons[i, j]);
          grph.FillPath(hexBrushes[i,j], fillPath);
          fillPath.CloseFigure();
        }
      }

      //縁取り線描画
      grph.DrawPath(PathPen, linePath);
    }

    /// <summary>Hex描画ブラシを設定する</summary>
    /// <param name="x">X座標</param>
    /// <param name="y">Y座標</param>
    /// <param name="brush">描画ブラシ</param>
    public void SetBrush(int x, int y, Brush brush)
    { hexBrushes[x, y] = brush; }

    public bool GetHexPosition(Point pt, out int xNumber, out int yNumber)
    {
      double powH = 0.25 * HexagonHeight * HexagonHeight;
      for (int i = 0; i < hexCenters.GetLength(0); i++)
        for (int j = 0; j < hexCenters.GetLength(1); j++)
        {
          int dx = pt.X - hexCenters[i, j].X;
          int dy = pt.Y - hexCenters[i, j].Y;
          if (dx * dx + dy * dy < powH)
          {
            xNumber = i;
            yNumber = j;
            return true;
          }
        }
      xNumber = yNumber = 0;
      return false;
    }

    public void DrawCircle(Graphics grph, int xNumber, int yNumber, Brush brush, Pen pen)
    {
      int d = HexagonHeight / 2;
      int d2 = d / 2;
      Point ctr = hexCenters[xNumber, yNumber];
      Rectangle rect = new Rectangle(ctr.X - d2, ctr.Y - d2, d, d);
      grph.FillEllipse(brush, rect);
      grph.DrawEllipse(pen, rect);
    }

    #endregion

  }

}
