﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WorkingDead
{
  /// <summary>A*アルゴリズム用のエージェントクラス</summary>
  public class AStarAgent : ImmutableAStarAgent
  {

    #region インスタンス変数・プロパティ

    /// <summary>地図</summary>
    protected AStarMap myMap;

    /// <summary>経路タイルリスト</summary>
    private List<tileWithScore> routeTiles = new List<tileWithScore>();

    /// <summary>スコア付タイル</summary>
    public tileWithScore[][] tiles;

    /// <summary>現在のタイルを設定・取得する</summary>
    public AStarMap.ImmutableTile CurrentTile { get; set; }

    /// <summary>移動速度[m/s]を設定・取得する</summary>
    public double Speed { get; set; } = 1.0;

    /// <summary>前移動の残り</summary>
    private double resLength = 0.0;

    /// <summary>到着済みか否か</summary>
    public bool Reached { get { return routeTiles.Count == 0; } }

    /// <summary>現在の方向を取得する</summary>
    public AStarMap.Direction CurrentDirection { get; protected set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="map">地図</param>
    /// <param name="initialPosition">初期位置</param>
    public AStarAgent(AStarMap map)
    {
      this.myMap = map;
      this.CurrentTile = myMap.GetTile(0, 0);
      this.tiles = new tileWithScore[map.XTiles][];
      this.myMap.AddAgent(this, 0, 0);

      for (int i = 0; i < tiles.Length; i++)
      {
        tiles[i] = new tileWithScore[map.YTiles];
        for (int j = 0; j < tiles[i].Length; j++)
          tiles[i][j] = new tileWithScore(myMap.GetTile(i, j));
      }
    }

    /// <summary>現在位置を設定する</summary>
    /// <param name="x">X座標</param>
    /// <param name="y">Y座標</param>
    public void SetPosition(int x, int y)
    {
      myMap.ResetAgent(CurrentTile.X, CurrentTile.Y);
      CurrentTile = myMap.GetTile(x, y);
      myMap.SetAgent(this, x, y);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>移動する</summary>
    /// <param name="sec">移動時間</param>
    /// <returns>目的地到着の真偽</returns>
    public bool Move(double sec)
    {
      //到着済の場合
      if (routeTiles.Count == 0) return true;

      //移動可能マス数
      double workLen = Speed * sec + resLength;
      int mTiles = (int)Math.Floor(workLen / myMap.TileSize);

      //1マスずつ移動
      for (int i = 0; i < mTiles; i++)
      {
        //到着した場合
        if (MoveOneTile())
        {
          resLength = 0;
          return true;
        }
      }

      //到着できなかった場合は残り移動可能長さを保存
      resLength = workLen % myMap.TileSize;
      return false;
    }

    /// <summary>1マス移動する</summary>
    /// <returns>目的地に到着した場合にはTrue</returns>
    public bool MoveOneTile()
    {
      //移動先に他エージェントがいる場合は移動不可
      if (routeTiles[0].tile.Agent != null) return false;

      //移動
      if (0 < routeTiles.Count)
      {
        AStarMap.ImmutableTile tl = routeTiles[0].tile;
        if (CurrentTile.UpTile != null && CurrentTile.UpTile.Equals(tl)) CurrentDirection = AStarMap.Direction.Up;
        else if (CurrentTile.RightUpTile != null && CurrentTile.RightUpTile.Equals(tl)) CurrentDirection = AStarMap.Direction.RightUp;
        else if (CurrentTile.RightDownTile != null && CurrentTile.RightDownTile.Equals(tl)) CurrentDirection = AStarMap.Direction.RightDown;
        else if(CurrentTile.DownTile != null && CurrentTile.DownTile.Equals(tl)) CurrentDirection = AStarMap.Direction.Down;
        else if (CurrentTile.LeftUpTile != null && CurrentTile.LeftUpTile.Equals(tl)) CurrentDirection = AStarMap.Direction.LeftUp;
        else if (CurrentTile.LeftDownTile != null && CurrentTile.LeftDownTile.Equals(tl)) CurrentDirection = AStarMap.Direction.LeftDown;

        SetPosition(tl.X, tl.Y);
        routeTiles.Remove(routeTiles[0]);
      }

      //最終タイルに到着
      return 0 == routeTiles.Count;
    }

    /// <summary>タイル通過のコストを得る（通過不可の場合は0）</summary>
    /// <param name="tileType">タイルの種別</param>
    /// <returns>タイル通過のコスト</returns>
    public virtual double GetCost(int tileType)
    { return tileType; }

    /// <summary>目標タイル（いすれか最短のもの）までのルートを探索する</summary>
    /// <param name="goals">目標タイル</param>
    /// <returns>ルートが見つかった場合にはTrue</returns>
    public bool FindRoute(List<AStarMap.ImmutableTile> goals)
    {
      routeTiles.Clear();
      if (goals.Contains(CurrentTile)) return true;

      if (goals.Count == 0) return false; //ゴールが存在しない

      List<tileWithScore> clTiles = new List<tileWithScore>();  //クローズタイル
      List<tileWithScore> opTiles = new List<tileWithScore>();  //オープンタイル

      //タイル探索状態を初期化
      for (int i = 0; i < tiles.Length; i++)
      {
        for (int j = 0; j < tiles[i].Length; j++)
        {
          tiles[i][j].status = tileWithScore.state.unprocessed;
          tiles[i][j].penalty = 0;
          tiles[i][j].parent = null;
        }
      }

      //ペナルティ付与
      SetPenalty();

      //現在位置のタイルをオープン
      tiles[CurrentTile.X][CurrentTile.Y].cost = 0;
      opTiles.Add(tiles[CurrentTile.X][CurrentTile.Y]);
      while (true)
      {
        //最小スコアタイルを取得
        tileWithScore minTile = null;
        double mScore = double.MaxValue;
        double mCost = double.MaxValue;
        foreach (tileWithScore tl in opTiles)
        {
          if (tl.score < mScore || (tl.score == mScore && tl.cost < mCost))
          {
            mScore = tl.score;
            mCost = tl.cost;
            minTile = tl;
          }
        }

        //最小スコアタイルがゴールタイルの場合は経路探索終了
        if (goals.Contains(minTile.tile))
        {
          routeTiles.Add(minTile);
          while (minTile.parent.parent != null)
          {
            minTile = minTile.parent;
            routeTiles.Add(minTile);
          }
          routeTiles.Reverse();
          return true;
        }

        //最小スコアタイルをクローズ
        clTiles.Add(minTile);
        opTiles.Remove(minTile);
        minTile.status = tileWithScore.state.closed;
        //隣接タイルのスコア計算
        List<tileWithScore> nbTiles = getNeighborTiles(minTile.tile);
        foreach (tileWithScore tl in nbTiles)
        {
          if (tl.status == tileWithScore.state.unprocessed)
          {
            double cost = GetCost(myMap.GetTile(tl.tile.X, tl.tile.Y).Type);
            //通過不可のタイルではない場合には評価
            if (cost != 0 && tl.tile.Agent == null)
            {
              //最小のヒューリスティクス
              int hls = tl.tile.GetDistance(goals[0]);
              for (int gl = 1; gl < goals.Count; gl++)
                hls = Math.Min(hls, tl.tile.GetDistance(goals[gl]));

              tl.cost = minTile.cost + cost + tl.penalty;
              tl.score = tl.cost + hls;
              opTiles.Add(tl);
              tl.status = tileWithScore.state.opened;
              tl.parent = minTile;
            }
          }
        }

        if (opTiles.Count == 0)
        {
          return false; //探索失敗
        }
      }
    }

    /// <summary>目標タイルまでのルートを探索する</summary>
    /// <param name="goal">目標タイル</param>
    /// <returns>ルートが見つかった場合にはTrue</returns>
    public bool FindRoute(AStarMap.ImmutableTile goal)
    {
      List<AStarMap.ImmutableTile> tiles = new List<AStarMap.ImmutableTile>();
      tiles.Add(goal);
      return FindRoute(tiles);
    }

    /// <summary>タイルにペナルティを与える</summary>
    protected virtual void SetPenalty() { }

    /// <summary>隣接タイルを取得する</summary>
    /// <param name="tile">計算対象タイル</param>
    /// <returns>隣接タイル</returns>
    private List<tileWithScore> getNeighborTiles(AStarMap.ImmutableTile tile)
    {
      List<tileWithScore> nTiles = new List<tileWithScore>();
      AStarMap.ImmutableTile tBuff;
      tBuff = tile.UpTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);
      tBuff = tile.DownTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);
      tBuff = tile.LeftUpTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);
      tBuff = tile.LeftDownTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);
      tBuff = tile.RightUpTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);
      tBuff = tile.RightDownTile;
      if (tBuff != null) nTiles.Add(tiles[tBuff.X][tBuff.Y]);

      return nTiles;
    }

    #endregion

    #region インナークラス定義

    /// <summary>スコア付のタイル</summary>
    public class tileWithScore
    {

      /// <summary>処理状況</summary>
      public enum state
      {
        unprocessed = 0,
        opened = 1,
        closed = 2
      }

      /// <summary>タイル</summary>
      public AStarMap.ImmutableTile tile;

      /// <summary>親タイル</summary>
      public tileWithScore parent { get; set; }

      /// <summary>処理状況</summary>
      public state status { get; set; }

      /// <summary>コスト（最小化すべき値）</summary>
      public double cost { get; set; }

      /// <summary>スコア（コスト+ヒューリスティック）</summary>
      public double score { get; set; }

      /// <summary>ペナルティ（他エージェントに起因する特殊なコスト）</summary>
      public double penalty { get; set; }

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="tile">タイル</param>
      public tileWithScore(AStarMap.ImmutableTile tile)
      {
        this.tile = tile;
        this.status = state.unprocessed;
      }
    }

    #endregion

  }

  #region 読み取り専用インターフェースの定義

  /// <summary>読み取り専用のA*アルゴリズム用のエージェントクラス</summary>
  public interface ImmutableAStarAgent
  {
    #region インスタンス変数・プロパティ

    /// <summary>現在のタイルを取得する</summary>
    AStarMap.ImmutableTile CurrentTile { get; }

    /// <summary>移動速度を取得する</summary>
    double Speed { get; }

    /// <summary>到着済みか否か</summary>
    bool Reached { get; }

    /// <summary>現在の方向を取得する</summary>
    AStarMap.Direction CurrentDirection { get; }

    /// <summary>タイル通過のコストを得る（通過不可の場合は0）</summary>
    /// <param name="tileType">タイルの種別</param>
    /// <returns>タイル通過のコスト</returns>
    double GetCost(int tileType);

    #endregion
  }

  #endregion


}
