﻿using System;
using System.Collections.Generic;

namespace WorkingDead
{
  /// <summary>A*アルゴリズム用の地図クラス</summary>
  /// <remarks>2次元の構造のみを処理する。コストなどの地形評価はAStarAgentに担わせる。</remarks>
  public class AStarMap
  {

    #region 列挙型定義

    /// <summary>方向</summary>
    public enum Direction
    {
      Up,
      RightUp,
      RightDown,
      Down,
      LeftDown,
      LeftUp
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>タイル</summary>
    private Tile[][] tiles;

    /// <summary>X方向タイル数を取得する</summary>
    public uint XTiles { get; private set; }

    /// <summary>Y方向タイル数を取得する</summary>
    public uint YTiles { get; private set; }

    /// <summary>タイルサイズを設定取得する（Default:0.4m）</summary>
    public double TileSize { get; set; } = 0.4;

    /// <summary>地図内のエージェントリスト</summary>
    private List<AStarAgent> agents = new List<AStarAgent>();

    /// <summary>地図内のエージェントを取得する</summary>
    public ImmutableAStarAgent[] Agents { get { return agents.ToArray(); } }

    /// <summary>現在の日時を設定・取得する</summary>
    public DateTime Now { get; set; }

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="xTiles">X方向タイル数</param>
    /// <param name="yTiles">Y方向タイル数</param>
    public AStarMap(uint xTiles, uint yTiles)
    {
      this.XTiles = xTiles;
      this.YTiles = yTiles;
      tiles = new Tile[XTiles][];
      for (int i = 0; i < tiles.Length; i++)
      {
        tiles[i] = new Tile[yTiles];
        for (int j = 0; j < tiles[i].Length; j++)
          tiles[i][j] = new Tile(this, i, j);
      }
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>タイルの種別を設定する</summary>
    /// <param name="types">タイル種別を表わす2次元配列[X座標,Y座標]</param>
    public void SetTileType(int[,] types)
    {
      for (int i = 0; i < tiles.Length; i++)
        for (int j = 0; j < tiles[i].Length; j++)
          tiles[i][j].Type = types[i, j];
    }

    /// <summary>タイルの種別を設定する</summary>
    /// <param name="x">X座標</param>
    /// <param name="y">Y座標</param>
    /// <param name="tType">タイル種別</param>
    public void SetTileType(int x, int y, int tType)
    { tiles[x][y].Type = tType; }

    /// <summary>タイル種別を取得する</summary>
    /// <param name="x">X座標</param>
    /// <param name="y">Y座標</param>
    /// <returns>タイル種別</returns>
    public ImmutableTile GetTile(int x, int y)
    { return tiles[x][y]; }

    /// <summary>エージェントを追加する</summary>
    /// <param name="agent">エージェント</param>
    /// <param name="x">X座標</param>
    /// <param name="y">Y座標</param>
    public void AddAgent(AStarAgent agent, int x, int y)
    {
      this.agents.Add(agent);
      tiles[x][y].Agent = agent;
      agent.SetPosition(x, y);
    }

    /// <summary>エージェントを削除する</summary>
    /// <param name="agent">エージェント</param>
    public void RemoveAgent(AStarAgent agent)
    {
      tiles[agent.CurrentTile.X][agent.CurrentTile.Y].Agent = null;
      this.agents.Remove(agent);
    }

    /// <summary>エージェントを動かす</summary>
    /// <param name="sec">時間[sec]</param>
    public void MoveAgents(double sec)
    {
      for (int i = 0; i < agents.Count; i++)
      {
        AStarAgent agt = agents[i];
        //現在のタイルを保存
        Tile tile = tiles[agt.CurrentTile.X][agt.CurrentTile.Y];
        agt.Move(sec);
        tile.Agent = null;
        tiles[agt.CurrentTile.X][agt.CurrentTile.Y].Agent = agt;
      }

      Now = Now.AddSeconds(sec);
    }

    public void SetAgent(ImmutableAStarAgent agent, int x, int y)
    { tiles[x][y].Agent = agent; }

    public void ResetAgent(int x, int y)
    { tiles[x][y].Agent = null; }

    #endregion

    #region staticメソッド

    /// <summary>相対的な方向を取得する</summary>
    /// <param name="baseDirection">基準となる方向</param>
    /// <param name="up">相対的上</param>
    /// <param name="down">相対的下</param>
    /// <param name="rUp">相対的右上</param>
    /// <param name="rDown">相対的右下</param>
    /// <param name="lUp">相対的左上</param>
    /// <param name="lDown">相対的左下</param>
    public static void GetRelativeDirection(Direction baseDirection,
      out Direction up, out Direction down, out Direction rUp, out Direction rDown, out Direction lUp, out Direction lDown)
    {
      switch (baseDirection)
      {
        //上
        case Direction.Up:
          up = Direction.Up;
          rUp = Direction.RightUp;
          rDown = Direction.RightDown;
          down = Direction.Down;
          lDown = Direction.LeftDown;
          lUp = Direction.LeftUp;
          return;
        //右上
        case Direction.RightUp:
          up = Direction.RightUp;
          rUp = Direction.RightDown;
          rDown = Direction.Down;
          down = Direction.LeftDown;
          lDown = Direction.LeftUp;
          lUp = Direction.Up;
          return;
        //右下
        case Direction.RightDown:
          up = Direction.RightDown;
          rUp = Direction.Down;
          rDown = Direction.LeftDown;
          down = Direction.LeftUp;
          lDown = Direction.Up;
          lUp = Direction.RightUp;
          return;
        //下
        case Direction.Down:
          up = Direction.Down;
          rUp = Direction.LeftDown;
          rDown = Direction.LeftUp;
          down = Direction.Up;
          lDown = Direction.RightUp;
          lUp = Direction.RightDown;
          return;
        //左下
        case Direction.LeftDown:
          up = Direction.LeftDown;
          rUp = Direction.LeftUp;
          rDown = Direction.Up;
          down = Direction.RightUp;
          lDown = Direction.RightDown;
          lUp = Direction.Down;
          return;
        //左上
        default:
          up = Direction.LeftUp;
          rUp = Direction.Up;
          rDown = Direction.RightUp;
          down = Direction.RightDown;
          lDown = Direction.Down;
          lUp = Direction.LeftDown;
          return;
      }
    }

    /// <summary>逆方向を取得する</summary>
    /// <param name="direction"></param>
    /// <returns></returns>
    public static Direction GetReverseDirection(Direction direction)
    {
      switch (direction)
      {
        case Direction.Up:
          return Direction.Down;
        case Direction.RightUp:
          return Direction.LeftDown;
        case Direction.RightDown:
          return Direction.LeftUp;
        case Direction.Down:
          return Direction.Up;
        case Direction.LeftDown:
          return Direction.RightUp;
        default:
          return Direction.RightDown;
      }
    }

    #endregion

    #region 位置情報を表わすインナークラス

    /// <summary>タイル</summary>
    public class Tile : ImmutableTile
    {

      #region インスタンス変数・プロパティ

      /// <summary>所属するマップ</summary>
      private AStarMap map;

      /// <summary>偶数タイルか否か</summary>
      private bool evenTile;

      /// <summary>距離計算用ダミーY座標</summary>
      private int y_dum;

      /// <summary>上方のタイルを取得する</summary>
      public ImmutableTile UpTile
      {
        get
        {
          if (Y != 0) return map.tiles[X][Y - 1];
          else return null;
        }
      }

      /// <summary>下方のタイルを取得する</summary>
      public ImmutableTile DownTile
      {
        get
        {
          if (Y != map.YTiles - 1) return map.tiles[X][Y + 1];
          else return null;
        }
      }

      /// <summary>左上方のタイルを取得する</summary>
      public ImmutableTile LeftUpTile
      {
        get
        {
          if (X == 0 || (evenTile && Y == 0)) return null;
          else return map.tiles[X - 1][evenTile ? Y - 1 : Y];
        }
      }

      /// <summary>左下方のタイルを取得する</summary>
      public ImmutableTile LeftDownTile
      {
        get
        {
          if (X == 0 || (!evenTile && Y == map.YTiles - 1)) return null;
          else return map.tiles[X - 1][evenTile ? Y : Y + 1];
        }
      }

      /// <summary>右上方のタイルを取得する</summary>
      public ImmutableTile RightUpTile
      {
        get
        {
          if (X == map.XTiles - 1 || (evenTile && Y == 0)) return null;
          else return map.tiles[X + 1][evenTile ? Y - 1 : Y];
        }
      }

      /// <summary>右下方のタイルを取得する</summary>
      public ImmutableTile RightDownTile
      {
        get
        {
          if (X == map.XTiles - 1 || (!evenTile && Y == map.YTiles - 1)) return null;
          else return map.tiles[X + 1][evenTile ? Y : Y + 1];
        }
      }

      /// <summary>X座標を取得する</summary>
      public int X { get; private set; }

      /// <summary>Y座標を取得する</summary>
      public int Y { get; private set; }

      /// <summary>タイル種別を設定・取得する</summary>
      public int Type { get; set; }

      /// <summary>エージェントを設定・取得する</summary>
      public ImmutableAStarAgent Agent { get; set; }

      #endregion

      #region コンストラクタ

      /// <summary>インスタンスを初期化する</summary>
      /// <param name="map">地図</param>
      /// <param name="x">X座標</param>
      /// <param name="y">Y座標</param>
      public Tile(AStarMap map, int x, int y)
      {
        this.map = map;
        this.X = x;
        this.Y = y;
        this.evenTile = (x % 2 == 0);
        //距離計算用ダミーY座標
        y_dum = (evenTile ? 2 * y : 1 + 2 * y);
      }

      #endregion

      #region インスタンスメソッド

      /// <summary>別タイルまでの距離を計算する（障害物無しの場合）</summary>
      /// <param name="targetTile">別タイル</param>
      /// <returns>別タイルまでの距離</returns>
      public int GetDistance(ImmutableTile targetTile)
      {
        int dx = Math.Abs(this.X - targetTile.X);
        int dy = Math.Abs(this.y_dum - ((Tile)targetTile).y_dum);
        return (dy <= dx ? dx : (dx + dy) / 2);
      }

      /// <summary>隣接タイルを取得する</summary>
      /// <returns>隣接タイル</returns>
      public List<ImmutableTile> GetNeighborTiles()
      { return GetSurroundingTiles(0); }

      /// <summary>周囲のタイルを取得する</summary>
      /// <param name="length">距離（タイル数）</param>
      /// <returns>周囲のタイル</returns>
      public List<ImmutableTile> GetSurroundingTiles(int length)
      {
        List<ImmutableTile> nTiles = new List<ImmutableTile>();

        //右上
        ImmutableTile tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.RightUpTile != null)
          {
            tBuff = tBuff.RightUpTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.DownTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        //右下
        tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.RightDownTile != null)
          {
            tBuff = tBuff.RightDownTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.LeftDownTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        //下
        tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.DownTile != null)
          {
            tBuff = tBuff.DownTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.LeftUpTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        //左下
        tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.LeftDownTile != null)
          {
            tBuff = tBuff.LeftDownTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.UpTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        //左上
        tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.LeftUpTile != null)
          {
            tBuff = tBuff.LeftUpTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.RightUpTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        //上
        tBuff = this;
        for (int i = 0; i <= length; i++)
        {
          if (tBuff.UpTile != null)
          {
            tBuff = tBuff.UpTile;
            if (i == length)
            {
              nTiles.Add(tBuff);
              for (int j = 0; j < length; j++)
              {
                tBuff = tBuff.RightDownTile;
                if (tBuff == null) break;
                nTiles.Add(tBuff);
              }
            }
          }
        }

        return nTiles;
      }

      /// <summary>タイルを取得する</summary>
      /// <param name="dir">方位</param>
      /// <returns>タイル</returns>
      public ImmutableTile GetTile(Direction dir)
      {
        switch (dir)
        {
          case Direction.Up:
            return UpTile;
          case Direction.RightUp:
            return RightUpTile;
          case Direction.RightDown:
            return RightDownTile;
          case Direction.Down:
            return DownTile;
          case Direction.LeftDown:
            return LeftDownTile;
          default:
            return LeftUpTile;
        }
      }

      #endregion

      #region 等価評価関数

      public override bool Equals(object obj)
      { return this.Equals(obj as ImmutableTile); }

      public bool Equals(ImmutableTile p)
      {
        // If parameter is null, return false.
        if (ReferenceEquals(p, null))
        {
          return false;
        }

        // Optimization for a common success case.
        if (ReferenceEquals(this, p))
        {
          return true;
        }

        // If run-time types are not exactly the same, return false.
        if (this.GetType() != p.GetType())
        {
          return false;
        }

        // Return true if the fields match.
        // Note that the base class is not invoked because it is
        // System.Object, which defines Equals as reference equality.
        return (X == p.X) && (Y == p.Y);
      }

      public override int GetHashCode()
      { return X * 0x00010000 + Y; }

      #endregion

    }

    #region 読み取り専用インタータイル

    /// <summary>読み取り専用タイル</summary>
    public interface ImmutableTile
    {

      /// <summary>上方のタイルを取得する</summary>
      ImmutableTile UpTile { get; }

      /// <summary>下方のタイルを取得する</summary>
      ImmutableTile DownTile { get; }

      /// <summary>左上方のタイルを取得する</summary>
      ImmutableTile LeftUpTile { get; }

      /// <summary>左下方のタイルを取得する</summary>
      ImmutableTile LeftDownTile { get; }

      /// <summary>右上方のタイルを取得する</summary>
      ImmutableTile RightUpTile { get; }

      /// <summary>右下方のタイルを取得する</summary>
      ImmutableTile RightDownTile { get; }

      /// <summary>X座標を取得する</summary>
      int X { get; }

      /// <summary>Y座標を取得する</summary>
      int Y { get; }

      /// <summary>タイル種別を取得する</summary>
      int Type { get; }

      /// <summary>エージェントを取得する</summary>
      ImmutableAStarAgent Agent { get; }

      /// <summary>別タイルまでの距離を計算する（障害物無しの場合）</summary>
      /// <param name="targetTile">別タイル</param>
      /// <returns>別タイルまでの距離</returns>
      int GetDistance(ImmutableTile targetTile);

      /// <summary>隣接タイルを取得する</summary>
      /// <returns>隣接タイル</returns>
      List<ImmutableTile> GetNeighborTiles();

      /// <summary>周囲のタイルを取得する</summary>
      /// <param name="length">距離（タイル数）</param>
      /// <returns>周囲のタイル</returns>
      List<ImmutableTile> GetSurroundingTiles(int length);

      /// <summary>タイルを取得する</summary>
      /// <param name="dir">方位</param>
      /// <returns>タイル</returns>
      ImmutableTile GetTile(Direction dir);
    }
  }

  #endregion

  #endregion

}
