﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WorkingDead
{
  public class LightingController
  {

    #region 列挙型定義

    public enum Grid
    {
      grd16,
      grd32,
      grd64
    }

    #endregion

    #region インスタンス変数・プロパティ

    /// <summary>地図</summary>
    private AStarMap myMap;

    /// <summary>人感センサ</summary>
    private HumanSensor[,] sensors;

    /// <summary>出力（0.0-1.0）</summary>
    private double[,] power;

    /// <summary>検出されたか否か</summary>
    private bool[,] detected;

    /// <summary>不在検出時刻</summary>
    private DateTime[,] dtUndetected;

    /// <summary>下限出力[-]を設定・取得する</summary>
    public double MinimumLightingRate { get; set; } = 0.0;

    /// <summary>フェードアウト時間[sec]を設定・取得する</summary>
    public int FadeoutTime { get; set; } = 3 * 60;

    /// <summary>不在検知からの待ち時間[sec]を設定・取得する</summary>
    public int WaitingTime { get; set; } = 5 * 60;

    /// <summary>X方向分割数を取得する</summary>
    public int XNumber { get; private set; }

    /// <summary>Y方向分割数を取得する</summary>
    public int YNumber { get; private set; }

    /// <summary>定格消費電力に対するオフセット比率[-]を取得する</summary>
    public double OffsetRate { get; private set; }

    /// <summary>オフセット効果計算用の傾き</summary>
    private double offB;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="aMap">地図</param>
    /// <param name="grid">グリッド種別</param>
    /// <param name="offsetRate">オフセット比率[-]</param>
    public LightingController(AStarMap aMap, Grid grid, double offsetRate)
    {
      this.myMap = aMap;
      this.OffsetRate = offsetRate;
      offB = (1.0 - offsetRate) / 1.0;

      int[] xStt, yStt;
      //3.2mグリッドで制御の場合
      if (grid == Grid.grd32)
      {
        xStt = new int[] { 0, 10, 19, 29, 38, 48, 57, 67, 76, 86, 95, 105 };
        yStt = new int[] { 0, 9, 18, 27, 36 };
      }
      //6.4mグリッドで制御の場合
      else if (grid == Grid.grd64)
      {
        xStt = new int[] { 0, 19, 38, 57, 76, 105 };
        yStt = new int[] { 0, 18, 36 };
      }
      //個別点灯
      else
      {
        xStt = new int[] { 0, 10, 19, 29, 38, 48, 57, 67, 76, 86, 95, 105 };
        yStt = new int[] { 0, 5, 9, 14, 18, 23, 27, 32, 36 };
      }

      XNumber = xStt.Length - 1;
      YNumber = yStt.Length - 1;
      sensors = new HumanSensor[YNumber, XNumber];
      power = new double[YNumber, XNumber];
      dtUndetected = new DateTime[YNumber, XNumber];
      detected = new bool[YNumber, XNumber];
      for (int i = 0; i < sensors.GetLength(0); i++)
        for (int j = 0; j < sensors.GetLength(1); j++)
          sensors[i, j] = new HumanSensor(myMap, xStt[j], yStt[i], xStt[j + 1] - xStt[j], yStt[i + 1] - yStt[i]);

      if (grid == Grid.grd32)
      {
        for (int i = 0; i < XNumber; i++)
          for (int j = 0; j < YNumber; j++)
            sensors[j, i].LampNumber = 2;
      }
      else if (grid == Grid.grd64)
      {
        for (int i = 0; i < XNumber; i++)
        {
          sensors[0, i].LampNumber = 8;
          sensors[1, i].LampNumber = 8;
        }
        sensors[0, 4].LampNumber = 12;
        sensors[1, 4].LampNumber = 12;
      }
      else
      {
        for (int i = 0; i < XNumber; i++)
          for (int j = 0; j < YNumber; j++)
            sensors[j, i].LampNumber = 1;
      }
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>状態を更新する</summary>
    public void Update()
    {
      //各ゾーンの制御を更新
      for (int i = 0; i < sensors.GetLength(0); i++)
      {
        for (int j = 0; j < sensors.GetLength(1); j++)
        {
          bool lstDtct = detected[i, j];
          detected[i, j] = sensors[i, j].FindOccupant();

          //検出時は100%出力
          if (detected[i, j]) power[i, j] = 1.0;
          //非検出時
          else
          {
            //検出→非検出に変化した場合は非検出開始時刻をリセット
            if (lstDtct) dtUndetected[i, j] = myMap.Now;

            double ctSec = (myMap.Now - dtUndetected[i, j]).TotalSeconds;

            //待ち時間中は出力100%
            if (ctSec < WaitingTime) power[i, j] = 1.0;
            //フェードアウト時間中は減光
            else if (ctSec < WaitingTime + FadeoutTime)
              power[i, j] = (1.0 - (ctSec - WaitingTime) / FadeoutTime) * (1.0 - MinimumLightingRate) + MinimumLightingRate;
            //フェードアウト時間経過後は最低出力
            else power[i, j] = MinimumLightingRate;
          }
        }
      }
    }

    /// <summary>出力を取得する</summary>
    /// <param name="x">X方向分割番号</param>
    /// <param name="y">Y方向分割番号</param>
    /// <returns>出力</returns>
    public double GetPower(int x, int y)
    { return power[y, x] * offB + OffsetRate; }

    /// <summary>出力を取得する</summary>
    /// <returns>出力</returns>
    public double GetPower()
    {
      double sum = 0;
      for (int i = 0; i < XNumber; i++)
        for (int j = 0; j < YNumber; j++)
          sum += GetPower(i, j) * sensors[j, i].LampNumber;
      return sum;
    }

    #endregion

  }
}
