﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WorkingDead
{
  /// <summary>オフィス執務者</summary>
  class OfficeOccupant : AStarAgent
  {

    #region 定数宣言

    /// <summary>計算時間数</summary>
    private const int CALCULATION_HOURS = 10;

    /// <summary>外出回数</summary>
    private const int OUTGOING_NUMBER = 10;

    /// <summary>打ち合わせ時間[min]</summary>
    private const int MEETING_MINUTES = 5;

    /// <summary>外出時間[min]</summary>
    private const int OUTGOING_MINUTES = 5;

    #endregion

    #region 列挙型定義

    /// <summary>執務者の状態</summary>
    public enum State
    {
      /// <summary>標準の状態</summary>
      None,
      /// <summary>外出のために移動中</summary>
      StartOutgoing,
      /// <summary>外出中</summary>
      Outgoing,
      /// <summary>打ち合わせのために移動中</summary>
      StartMeeting,
      /// <summary>打ち合わせ中</summary>
      Meeting
    }

    #endregion

    #region インスタンス変数・プロパティ
    
    /// <summary>打ち合わせ相手</summary>
    private ImmutableAStarAgent meetingTarget = null;

    /// <summary>打ち合わせ開始時刻</summary>
    private DateTime startTime;

    /// <summary>打ち合わせの深さ（0: 隣接, 1: 1タイル間を置く、2: 2タイル間を置く...）</summary>
    private int meetingDepth;

    /// <summary>乱数製造機</summary>
    private Random rnd;

    /// <summary>状態を取得する</summary>
    public State Status { get; private set; }

    /// <summary>自席のタイルを取得する</summary>
    public AStarMap.ImmutableTile SeatTile { get; private set; }

    /// <summary>着席時の方位を設定・取得する</summary>
    public AStarMap.Direction SeatDirection { get; private set; }

    /// <summary>打ち合わせを行う時刻のリスト</summary>
    public List<DateTime> MeetingList = new List<DateTime>();

    /// <summary>外出する時刻のリスト</summary>
    public List<DateTime> OutgoingList = new List<DateTime>();

    /// <summary>話しかけられて動けない状態か否か</summary>
    public bool IsLocked { get { return LockingAgent != null; } }

    /// <summary>外出用ドア</summary>
    public List<AStarMap.ImmutableTile> Doors = new List<AStarMap.ImmutableTile>();

    /// <summary>外出した際に使ったドア</summary>
    private AStarMap.ImmutableTile myDoor;

    /// <summary>自分に話しかけている相手を設定・取得する</summary>
    public ImmutableAStarAgent LockingAgent { get; set; } = null;

    /// <summary>打ち合わせ回数を設定・取得する</summary>
    public int MeetingNumber { get; set; } = 5;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="map"></param>
    /// <param name="seat">自席のタイル</param>
    public OfficeOccupant(AStarMap map, AStarMap.ImmutableTile seat, AStarMap.Direction seatDirection, int randomSeed)
      : base(map)
    {
      this.SeatTile = seat;
      this.SeatDirection = seatDirection;
      rnd = new Random(randomSeed);
    }

    #endregion

    #region インスタンスメソッド

    /// <summary>タイル通過のコストを得る（通過不可の場合は0）</summary>
    /// <param name="tileType">タイルの種別</param>
    /// <returns>タイル通過のコスト</returns>
    public override double GetCost(int tileType)
    {
      switch (tileType)
      {
        case 0: return 0;   //通行不可
        case 1: return 1.5;  //通路外
        case 2: return 1.0;   //通路
        case 3: return 1.5;   //席
        case 4: return 1.0;   //扉
        default: return 0;  //通行不可
      }
    }

    /// <summary>ペナルティを付与する</summary>
    protected override void SetPenalty()
    {
      for (int i = 0; i < tiles.Length; i++)
        for (int j = 0; j < tiles[i].Length; j++)
          tiles[i][j].penalty = 0;

      //他の執務者のそばは通りたくない!
      ImmutableAStarAgent[] agents = myMap.Agents;
      foreach (ImmutableAStarAgent agt in agents)
      {
        AStarMap.Direction up, down, rUp, rDown, lUp, lDown;
        AStarMap.GetRelativeDirection
          (agt.CurrentDirection, out up, out down, out rUp, out rDown, out lUp, out lDown);

        AStarMap.ImmutableTile tl1, tl2, tl3;
        if ((tl1 = agt.CurrentTile.GetTile(up)) != null)
        {
          tiles[tl1.X][tl1.Y].penalty += 1.0; //でっかく
          if ((tl2 = tl1.GetTile(lUp)) != null) tiles[tl2.X][tl2.Y].penalty += 0.4;
          if ((tl2 = tl1.GetTile(rUp)) != null) tiles[tl2.X][tl2.Y].penalty += 0.4;

          if ((tl2 = tl1.GetTile(up)) != null)
          {
            tiles[tl2.X][tl2.Y].penalty += 0.6;
            if ((tl3 = tl2.GetTile(lUp)) != null) tiles[tl3.X][tl3.Y].penalty += 0.2;
            if ((tl3 = tl2.GetTile(rUp)) != null) tiles[tl3.X][tl3.Y].penalty += 0.2;
            if ((tl3 = tl2.GetTile(up)) != null) tiles[tl3.X][tl3.Y].penalty += 0.4;
          }
        }

        if ((tl1 = agt.CurrentTile.GetTile(rUp)) != null)
        {
          tiles[tl1.X][tl1.Y].penalty += 1.0;
          if ((tl2 = tl1.GetTile(rUp)) != null) tiles[tl2.X][tl2.Y].penalty += 0.2;
        }

        if ((tl1 = agt.CurrentTile.GetTile(lUp)) != null)
        {
          tiles[tl1.X][tl1.Y].penalty += 1.0;
          if ((tl2 = tl1.GetTile(lUp)) != null) tiles[tl2.X][tl2.Y].penalty += 0.2;
        }

        if ((tl1 = agt.CurrentTile.GetTile(rDown)) != null)
          tiles[tl1.X][tl1.Y].penalty += 0.8;

        if ((tl1 = agt.CurrentTile.GetTile(down)) != null)
          tiles[tl1.X][tl1.Y].penalty += 0.8;

        if ((tl1 = agt.CurrentTile.GetTile(lDown)) != null)
          tiles[tl1.X][tl1.Y].penalty += 0.8;
      }
    }

    /// <summary>打ち合わせ・外出のスケジュールをランダムに生成する</summary>
    public void MakeSchedule()
    {
      //総秒数
      int sec = CALCULATION_HOURS * 60 * 60;

      //打ち合わせリストの作成
      for (int i = 0; i < MeetingNumber; i++)
        MeetingList.Add(myMap.Now.AddSeconds(rnd.Next(sec)));
      MeetingList.Sort();

      //外出リストの作成
      for (int i = 0; i < OUTGOING_NUMBER; i++)
        OutgoingList.Add(myMap.Now.AddSeconds(rnd.Next(sec)));
      OutgoingList.Sort();
    }

    /// <summary>状態を更新する</summary>
    public void Update()
    {
      //話しかけられている時には相手の方向を向いて何もできない
      if (IsLocked)
      {
        CurrentDirection = AStarMap.GetReverseDirection(LockingAgent.CurrentDirection);
        return;
      }

      //手持ち無沙汰の場合は行動判定
      if (Status == State.None)
      {
        //打ち合わせをする場合
        if (MeetingList.Count != 0 && MeetingList[0] < myMap.Now)
        {
          MeetingList.Remove(MeetingList[0]);
          Status = State.StartMeeting;

          //打ち合わせ相手をランダムに特定（自分以外）
          ImmutableAStarAgent[] agent = myMap.Agents;
          if (2 < agent.Length)
          {
            while ((meetingTarget = agent[rnd.Next(agent.Length)]) != this)
            {
              meetingDepth = rnd.Next(3);
              break;
            }
          }
        }

        //外出する場合
        else if (OutgoingList.Count != 0 && OutgoingList[0] < myMap.Now)
        {
          OutgoingList.Remove(OutgoingList[0]);
          Status = State.StartOutgoing;
        }

        //なにもしない場合にはひたすら自席を目指す//到着したら席の向き
        else if (FindRoute(SeatTile) && Reached) CurrentDirection = SeatDirection;

      }

      //打ち合わせに向けて移動中の場合
      if (Status == State.StartMeeting)
      {
        OfficeOccupant tgt = (OfficeOccupant)meetingTarget;

        //相手が忙しそうにしてたら帰る
        if (!myMap.Agents.Contains(tgt) || tgt.Status == State.Meeting || tgt.Status == State.StartMeeting)
          Status = State.None;
        //相手にたどり着いた場合
        else if (FindRoute(meetingTarget.CurrentTile.GetSurroundingTiles(meetingDepth)) && Reached)
        {
          //相手が暇そうならばロックして打ち合わせ開始
          if (!tgt.IsLocked && tgt.Status == State.None)
          {
            tgt.LockingAgent = this;
            Status = State.Meeting;
            startTime = myMap.Now;
          }
          //相手が忙しそうならばトボトボ帰る
          else Status = State.None;
        }
      }

      //打ち合わせ継続中の場合
      else if (Status == State.Meeting)
      {
        //十分に打ち合わせたらロックを解除
        if (startTime.AddMinutes(MEETING_MINUTES) < myMap.Now)
        {
          ((OfficeOccupant)meetingTarget).LockingAgent = null;
          Status = State.None;
        }
      }

      //外出に向けて移動中の場合
      else if (Status == State.StartOutgoing)
      {
        //ドアにたどり着いた場合
        if (FindRoute(Doors) && Reached)
        {
          myDoor = CurrentTile; //外出時のドアを記録
          myMap.RemoveAgent(this);  //自分を削除
          Status = State.Outgoing;  //状態変更+外出時間計測開始
          startTime = myMap.Now;
        }
      }

      //外出中の場合
      else if (Status == State.Outgoing)
      {
        //十分に暇を潰したら戻る
        if (startTime.AddMinutes(OUTGOING_MINUTES) < myMap.Now)
        {
          myMap.AddAgent(this, myDoor.X, myDoor.Y);
          Status = State.None;
        }
      }

    }

    #endregion
    
  }
}
